#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# All bundle commands are here
#

import os
import logging
import tempfile
import shutil
import glob

import Blocks.Packaging as BP
from Blocks.Packaging.DependencyProcessors.RomPatchProcessor import RomPatchDependencyProcessor
from Blocks.Packaging.DependencyProcessors.RaptorDependencyProcessor import DotDeeDependencyProcessor
from Blocks.Packaging.DependencyProcessors.DefaultProcessors import BuildDataDependencyProcessor
from Blocks.Packaging.DataSources.LinkInfoToBuildData import LinkInfo
from Blocks.debfile import DebFile, DebError
import Blocks.gpg as gpg

from cmdlinecommand import CmdLineCommand, CommandOptionParser
import utils
import elementtree as ET
from generalexceptions import InvalidFileInput, InvalidUserInput, GenericError, ExternalProcessError
from UserConfiguration.UserConfiguration import UserConfiguration
from cmdlineutils import Dpkg, CmdLineUtil
import comps

GPG_PASSPHRASEFILE_ENVVAR = "BLOCKS_GPG_PASSPHRASEFILE"

class BlocksDependency(object):
    ''' Blocks workspace dependency resolver '''
    def __init__(self, root):
        self.dpkg = None
        paths = self._getWorkspacePaths(root)
        if paths:
            self.dpkg = Dpkg(*paths)

    def getPackageName(self, path):
        if self.dpkg:
            try:
                output = self.dpkg("-S %s" % path, True)
            except ExternalProcessError:
                pass
            else:
                packagename = output.split(":", 1)[0]
                return packagename

    @staticmethod
    def _getWorkspacePaths(path):
        config = UserConfiguration()
        workspacePaths = [(config.getWorkspacePath(wsid), config.getWorkspaceMetadataPath(wsid))
                          for wsid in config.getWorkspaces()]
        for wsroot, metapath in workspacePaths:
            if utils.pathInside(wsroot, path):
                return (wsroot, metapath)

class BundleCommand(CmdLineCommand):
    ''' Bundle commands '''
    def cmd_create(self, xmlfile):
        ''' Create bundle(s) '''
        logFormat, logTimeFormat = utils.getConsoleLogFormat()
        BP.Logging.setupLogHandling(utils.loglevel, logFormat=logFormat, logTimeFormat=logTimeFormat)
        targetroot = os.path.abspath(self.cmd_options.workspaceroot)
        sourceroot = os.path.abspath(self.cmd_options.sourceroot or targetroot)

        file_directives_path, pkg_directives_path = self.getDirectivePaths(self.cmd_options.directives)
        storage = BP.OneoffStorage(self.cmd_options.metadir)
        utils.createDir(self.cmd_options.outputdir)
        packager = BP.Packager(storage=storage, packageOutputDirectory=self.cmd_options.outputdir,
                               targetRules=file_directives_path, sourceRules=file_directives_path,
                               directives=pkg_directives_path, keepGoing=False)

        # Dependency processors
        packager.globalFileMap.registerExternalSource(BlocksDependency(sourceroot))
        packager.addProcessor(RomPatchDependencyProcessor)
        packager.addProcessor(BuildDataDependencyProcessor, None)
        if self.cmd_options.sbs_dep_log:
            packager.addProcessor(DotDeeDependencyProcessor, {"directivesPath": pkg_directives_path})

        try:
            tree = ET.MyElementTree(file=xmlfile)
        except SyntaxError, ex:
            raise GenericError("Bundle XML-file %s" % ex)

        for bundle in tree.getiterator("bundle"):
            component = BP.PlainBuildData()
            try:
                component.setTargetRoot(targetroot)
                component.setSourceRoot(sourceroot)
                component.setComponentName(bundle.findtext("name"))
                component.setComponentVersion(bundle.findtext("version"))
                for attribute in bundle.findall("meta"):
                    attrName = attribute.get("name")
                    if not attrName:
                        raise ValueError("Name for attribute not defined")
                    component.attributes[attrName] = attribute.text
            except ValueError, ex:
                raise InvalidFileInput("XML-file error: %s" % ex)
            targets = bundle.find("targets")
            if targets:
                targetFiles = [utils.relativePath(t.text, targetroot) for t in targets.findall("target")]
                logging.log(logging.DEBUG2, "Target files: %s", targetFiles)
                component.addTargetFiles(targetFiles)
            sources = bundle.find("sources")
            if sources:
                sourceFiles = [utils.relativePath(t.text, sourceroot) for t in sources.findall("source")]
                logging.log(logging.DEBUG2, "Source files: %s", sourceFiles)
                component.addSourceFiles(sourceFiles)

            # Dependency processing
            component.dependencyData["DotDeeDependencyProcessor"] = {}
            if self.cmd_options.sbs_build_dir:
                component.dependencyData["DotDeeDependencyProcessor"]["buildDir"] = self.cmd_options.sbs_build_dir
            bldinfPaths = [e.text for e in bundle.findall("bldinf")]
            component.dependencyData["DotDeeDependencyProcessor"]["infs"] = bldinfPaths
            logging.debug("Bldinfs: %s", bldinfPaths)
            if self.cmd_options.sbs_dep_log:
                linkinfo = LinkInfo(self.cmd_options.sbs_dep_log)
                if bldinfPaths:
                    for infpath in bldinfPaths:
                        linkinfo.addDependencies(component, os.path.abspath(infpath))
                else:
                    linkinfo.addDependencies(component)
                component.addNonstandardInterfaces(linkinfo.getNonstandardInterfaces())
            packager.addComponent(component)
        try:
            packager.wait()
        except BP.PackagingError, ex:
            raise GenericError("Packaging error: %s" % ex)

    @staticmethod
    def getDirectivePaths(name):
        paths = [None, None]
        if not name:
            return paths
        for path in (utils.getMetaPath(), utils.getInstallDir()):
            directive_path = os.path.join(path, "conf", "directives", name)
            if os.path.isdir(directive_path):
                file_directives_path = os.path.join(directive_path, "file-directives.xml")
                pkg_directives_path = os.path.join(directive_path, "pkg-directives.xml")
                if os.path.isfile(file_directives_path):
                    paths[0] = file_directives_path
                if os.path.isfile(pkg_directives_path):
                    paths[1] = pkg_directives_path
            if not paths[0] and not paths[1]:
                logging.debug("Directive '%s' not found from '%s'", name, directive_path)
            else:
                break
        if not paths[0] and not paths[1]:
            raise InvalidUserInput("Directive '%s' not found" % name)
        logging.debug("File directive: %s, Package directive: %s", paths[0], paths[1])
        return paths

    cmd_create._optionparser = CommandOptionParser()
    cmd_create._optionparser.add_option("-w", "--workspaceroot", metavar="DIR", help="Target files root [Default: Current directory]")
    cmd_create._optionparser.add_option("-s", "--sourceroot", metavar="DIR", help="Source files root [Default: Same as workspace root]")
    cmd_create._optionparser.add_option("-m", "--metadir", metavar="DIR", help="Directory for meta-data storage [Default: Temp directory]")
    cmd_create._optionparser.add_option("-o", "--outputdir", metavar="DIR", help="Output directory [Default: '%default']")
    cmd_create._optionparser.add_option("-l", "--sbs-dep-log", metavar="FILE", help="Blocks SBS-plugin log file")
    cmd_create._optionparser.add_option("-b", "--sbs-build-dir", metavar="DIR", help="SBS build directory")
    cmd_create._optionparser.add_option("-d", "--directives", metavar="name", help="Override default directives (e.g. single-bundle)")
    cmd_create._optionparser.set_defaults(workspaceroot=".", outputdir="bundles")

    def cmd_create_xml(self, xmlfile):
        ''' Create XML-file for use with create command '''
        if self.cmd_options.append:
            tree = ET.MyElementTree(file=xmlfile)
        else:
            tree = ET.MyElementTree(ET.Element("bundles"))

        bundle = ET.SubElement(tree.getroot(), "bundle")
        ET.SubElement(bundle, "name").text = self.cmd_options.name
        ET.SubElement(bundle, "version").text = self.cmd_options.version
        for attr in self.cmd_options.attribute:
            key, _, value = attr.partition("=")
            if value:
                meta = ET.SubElement(bundle, "meta")
                meta.text = value
                meta.set("name", key)

        targets = ET.SubElement(bundle, "targets")
        sources = ET.SubElement(bundle, "sources")
        for path in utils.getFileLines(self.cmd_options.targetfiles_file):
            ET.SubElement(targets, "target").text = path
            utils.warnIfFileNotFound(path)
        for path in utils.getFileLines(self.cmd_options.sourcefiles_file):
            ET.SubElement(sources, "source").text = path
            utils.warnIfFileNotFound(path)

        for name in self.getFiles(self.cmd_options.target):
            ET.SubElement(targets, "target").text = name
        for name in self.getFiles(self.cmd_options.source):
            ET.SubElement(sources, "source").text = name

        tree.write(xmlfile)

    cmd_create_xml._optionparser = CommandOptionParser()
    cmd_create_xml._optionparser.add_option("-n", "--name", help="Component name [Default: %default]")
    cmd_create_xml._optionparser.add_option("-v", "--version", help="Component version [Default: %default]")
    cmd_create_xml._optionparser.add_option("-t", "--target", action="append", metavar="FILE/DIR", help="Add target file/dir")
    cmd_create_xml._optionparser.add_option("-s", "--source", action="append", metavar="FILE/DIR", help="Add source file/dir")
    cmd_create_xml._optionparser.add_option("-T", "--targetfiles-file", metavar="FILE", help="Specify file which contains target files")
    cmd_create_xml._optionparser.add_option("-S", "--sourcefiles-file", metavar="FILE", help="Specify file which contains source files")
    cmd_create_xml._optionparser.add_option("-a", "--attribute", action="append", metavar="NAME=VALUE", help="Add bundle attribute")
    cmd_create_xml._optionparser.add_option("--append", action="store_true",
                                            help="Append component to XML-file instead of overwriting it [Default: %default]")
    cmd_create_xml._optionparser.set_defaults(name="bundle", version="1.0.0", target=[], source=[], attribute=[], append=False)

    @staticmethod
    def getFiles(pathlist):
        '''
        Get list of files from pathlist.
        Add all files from list and add all files from paths recursively in pathlist
        '''
        files = []
        for path in pathlist:
            path = os.path.abspath(path)
            if os.path.isdir(path):
                files.extend(utils.getFilesRecursive(path))
            else:
                files.append(path)
                utils.warnIfFileNotFound(path)
        return files

    def cmd_compare(self, bundle1, bundle2):
        ''' Compare two bundles for differences '''
        for name in (bundle1, bundle2):
            if not os.path.isfile(name):
                raise IOError("Bundle file '%s' not found" % name)

##        if filecmp.cmp(bundle1, bundle2, False):
##            print "Files are identical"
##            return

        try:
            deb1 = DebFile(bundle1)
            deb2 = DebFile(bundle2)
        except DebError, ex:
            raise IOError(str(ex))
        metadiff, filediff = deb1.compare(deb2)
        if filediff.changed:
            tempdir = tempfile.mkdtemp(prefix="bundle-compare")
            try:
                bundledir1 = os.path.join(tempdir, "bundle1")
                bundledir2 = os.path.join(tempdir, "bundle2")
                os.mkdir(bundledir1)
                os.mkdir(bundledir2)
                for filename in filediff.changed:
                    deb1.extractData(filename, bundledir1)
                    deb2.extractData(filename, bundledir2)
                changedfiles = self._getEffectivelyChangedFiles(bundledir1, bundledir2, filediff.changed)
                filediff = filediff._replace(changed=changedfiles)
            finally:
                shutil.rmtree(tempdir)

        metadataDifferent = any(metadiff)
        filesDifferent = any(filediff)
        if metadataDifferent or filesDifferent:
            if metadataDifferent:
                print "Metadata differences:"
                for name in sorted(metadiff.new):
                    print "N %s: %s" % (name, deb2.metadata[name])
                for name in sorted(metadiff.removed):
                    print "R %s: %s" % (name, deb1.metadata[name])
                for name in sorted(metadiff.changed):
                    print "M %s: %s -> %s" % (name, deb1.metadata[name], deb2.metadata[name])
                if filesDifferent:
                    print
            if filesDifferent:
                print "File differences:"
                for name in sorted(filediff.new):
                    print "N %s" % name
                for name in sorted(filediff.removed):
                    print "R %s" % name
                for name in sorted(filediff.changed):
                    print "M %s" % name
        else:
            print "Bundle content identical"

    @staticmethod
    def _getEffectivelyChangedFiles(path1, path2, files):
        diffFiles = []

        bd1 = BP.PlainBuildData()
        bd1.setTargetRoot(path1)
        bd1.addTargetFiles(files)
        bd1.setComponentName("temp")

        bd2 = BP.PlainBuildData()
        bd2.setTargetRoot(path2)
        bd2.addTargetFiles(files)
        bd2.setComponentName("temp")

        builder = BP.ComponentBuilder.ComponentBuilder()
        c1 = builder.createComponent(BP.OneoffStorage(), bd1)
        c2 = builder.createComponent(BP.OneoffStorage(), bd2)
        differentPackages = c1.diff(c2)[3]
        for package in differentPackages:
            p1 = c1.getPackage(package)
            p2 = c2.getPackage(package)
            diffFiles.extend(p1.diff(p2)[3])

        return diffFiles

    @staticmethod
    def _validateSignType(signtype):
        if len(signtype) > 10:
            raise InvalidUserInput("Signature type length must be between 1 and 10 characters")
        if not signtype.isalpha():
            raise InvalidUserInput("Signature type must contain only alphabets")

    def cmd_sign(self, bundle1, *bundle):
        ''' Sign bundle(s) '''
        self._validateSignType(self.cmd_options.type)
        for bundle in [bundle1] + list(bundle):
            deb = None
            try:
                deb = DebFile(bundle, "a")
                if self.cmd_options.clear:
                    try:
                        deb.removeSignature()
                    except DebError, ex:
                        logging.warning("No signatures to remove")
                elif deb.signatureExists(self.cmd_options.type):
                    deb.removeSignature(self.cmd_options.type)
                    logging.warning("Overwriting existing signature type '%s' on bundle '%s'", self.cmd_options.type, bundle)
                passphraseFile = os.environ.get(GPG_PASSPHRASEFILE_ENVVAR)
                deb.addSignature(self.cmd_options.type, gpgPassfile=passphraseFile)
            except DebError, ex:
                raise GenericError("Signing failed on bundle '%s': %s" % (bundle, ex.error))
            if deb:
                deb.close()

    cmd_sign._optionparser = CommandOptionParser()
    cmd_sign._optionparser.add_option("-t", "--type", help="Signature type [Default: %default]")
    cmd_sign._optionparser.add_option("-c", "--clear", action="store_true", help="Clear all existing signatures [Default: %default]")
    cmd_sign._optionparser.set_defaults(type="origin", clear=False)

    def cmd_sign_remove(self, bundle1, *bundle):
        ''' Remove signatures from bundle(s) '''
        self._validateSignType(self.cmd_options.type)
        if self.cmd_options.type == "all":
            self.cmd_options.type = None
        for bundle in [bundle1] + list(bundle):
            deb = None
            try:
                deb = DebFile(bundle, "a")
                deb.removeSignature(self.cmd_options.type)
            except DebError, ex:
                raise GenericError("Signature removal failed on bundle '%s': %s" % (bundle, ex.error))
            if deb:
                deb.close()

    cmd_sign_remove._optionparser = CommandOptionParser()
    cmd_sign_remove._optionparser.add_option("-t", "--type", help="Signature type [Default: %default]")
    cmd_sign_remove._optionparser.set_defaults(type="all")

    def cmd_verify(self, bundle1, *bundle):
        ''' Verify signatures in bundle(s) '''
        self._validateSignType(self.cmd_options.type)
        for bundle in [bundle1] + list(bundle):
            if self.cmd_options.type == "all":
                self.cmd_options.type = None
            status = []
            deb = None
            try:
                deb = DebFile(bundle, "r")
                status = deb.verifySignature(self.cmd_options.type)
            except DebError, ex:
                raise GenericError("Signature verification failed on bundle '%s': %s" % (bundle, ex.error))
            if deb:
                deb.close()
            for signTypeStatus in status:
                if signTypeStatus.status.code == gpg.GpgStatusCode.VERIFIED:
                    print "Verified signature type '%s' from '%s' in bundle '%s'." % (signTypeStatus.type, signTypeStatus.status.info.name, bundle)
                elif signTypeStatus.status.code == gpg.GpgStatusCode.BADSIG:
                    raise GenericError("Bad signature with type '%s' in bundle '%s'!" % (signTypeStatus.type, bundle))
                elif signTypeStatus.status.code == gpg.GpgStatusCode.NO_PUBKEY:
                    raise GenericError("Signature with type '%s' in bundle '%s' can't be verified because public key is not available." % (signTypeStatus.type, bundle))
                elif signTypeStatus.status.code == gpg.GpgStatusCode.KEYEXPIRED:
                    raise GenericError("Public key for bundle '%s' with signature type '%s' has been expired." % (bundle, signTypeStatus.type))
                elif signTypeStatus.status.code == gpg.GpgStatusCode.REVKEYSIG:
                    raise GenericError("Public key for bundle '%s' with signature type '%s' has been revoked." % (bundle, signTypeStatus.type))
                else:
                    raise GenericError("Unknown verification error with signature type '%s' for bundle '%s'" % (signTypeStatus.type, bundle))

    cmd_verify._optionparser = CommandOptionParser()
    cmd_verify._optionparser.add_option("-t", "--type", help="Signature type [Default: %default]")
    cmd_verify._optionparser.set_defaults(type="all")

    def cmd_sign_list(self, bundle):
        ''' List signature types in a bundle '''
        try:
            deb = DebFile(bundle, "r")
            signs = deb.getSignatures()
            if signs:
                print "Signature types:"
                print "\n".join(signs)
            else:
                print "No signatures found"
        except DebError, ex:
            raise IOError(ex.error)

    TEMPLATE_FILE = os.path.join(utils.getInstallDir(), "conf", "comps_template.xml")
    def cmd_create_repo(self, path):
        ''' Create/update repository index '''
        oldDir = os.getcwd()
        try:
            os.chdir(path)
        except OSError:
            raise InvalidUserInput("Path '%s' not found" % path)
        if not any(glob.iglob("*.deb")):
            logging.warning("Path '%s' does not contain any bundles(*.deb).\nCreating empty repository.", path)

        ftparchive = CmdLineUtil("apt-ftparchive")
        ftparchive("packages . > Packages")
        if self.cmd_options.sign:
            ftparchive("release . > Release")
            passphraseFile = os.environ.get(GPG_PASSPHRASEFILE_ENVVAR)
            try:
                gpg.sign("Release", "Release.gpg", passfile=passphraseFile)
            except gpg.GpgError, ex:
                raise InvalidUserInput(str(ex.output))

        os.chdir(oldDir)
        if not self.cmd_options.no_comps:
            self._createComps(path, self.cmd_options.comps_template)

    cmd_create_repo._optionparser = CommandOptionParser()
    cmd_create_repo._optionparser.add_option("-s", "--sign", action="store_true", help="Sign repository [Default: %default]")
    cmd_create_repo._optionparser.add_option("-t", "--comps-template", metavar="FILE", help="comps.xml template [Default: %default]")
    cmd_create_repo._optionparser.add_option("--no-comps", action="store_true", help="Do not generate group information (comps.xml) [Default: %default]")
    cmd_create_repo._optionparser.set_defaults(sign=False, comps_template=TEMPLATE_FILE, no_comps=False)

    @staticmethod
    def _createComps(path, template):
        try:
            compsTemplate = comps.CompsTemplate(template)
        except SyntaxError, ex:
            raise InvalidFileInput("Error in comps template:\n%s" % ex)
        except IOError, ex:
            raise IOError("Comps template: %s" % ex)

        for bundle in glob.glob(os.path.join(path, "*.deb")):
            try:
                deb = DebFile(bundle)
            except DebError, ex:
                raise InvalidFileInput("Error on bundle '%s': %s" % (bundle, ex))
            compsTemplate.applyRule(deb.metadata)
            deb.close()
        absPath = os.path.join(os.path.abspath(path), "comps.xml")
        compsTemplate.create(absPath)