#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Workspace diffing
#

import os
from hashlib import md5

from generalexceptions import InvalidUserInput
import utils

class DefaultOutput(object):
    ''' Default output class for WorkspaceDiff '''

    def __init__(self, differ):
        self.differ = differ
        self.currentBundle = ""
        self._clear()

    def _clear(self):
        self.removed_files = []
        self.modified_files = []
        self.unmodified_files = []

    def _bundlePrint(self, bundle):
        print "%s%s%s %s" % ("M" if self.modified_files else " ",
                             "R" if self.removed_files else " ",
                             "U" if self.unmodified_files else " ",
                             bundle)

    def _filePrint(self, bundle):
        print "%s:" % bundle
        if self.modified_files:
            print "\n".join("  M " + n for n in self.modified_files)
        if self.removed_files:
            print "\n".join("  R " + n for n in self.removed_files)
        if self.unmodified_files:
            print "\n".join("  U " + n for n in self.unmodified_files)

    def _print(self, bundle):
        if self.modified_files or self.removed_files or self.unmodified_files:
            if self.differ.bundlediff:
                self._bundlePrint(bundle)
            else:
                self._filePrint(bundle)

    def __call__(self, status, bundle, name):
        if status == "END":
            self._print(self.currentBundle)
            if self.differ.new:
                print
            return True

        if not self.currentBundle:
            self.currentBundle = bundle
        if bundle != self.currentBundle and bundle:
            self._print(self.currentBundle)
            self.currentBundle = bundle
            self._clear()

        if self.differ.bundlediff:
            if status == "M":
                self.modified_files = True
            elif status == "R":
                self.removed_files = True
            elif status == "U":
                self.unmodified_files = True
            # Stop scanning this bundle if all needed flags found
            if (self.modified_files == self.differ.modified and
                self.removed_files == self.differ.removed and
                self.unmodified_files == self.differ.unmodified):
                return True
        else:
            if status == "N":
                print "N", name
            else:
                {"M": self.modified_files, "R": self.removed_files, "U": self.unmodified_files}[status].append(name)

class WorkspaceDiff(object):
    ''' Find changes in workspace '''

    def __init__(self, wsPath, wsMetaDir, output=DefaultOutput):
        self.wsPath = wsPath
        self.wsMetaDir = wsMetaDir
        self.output = output(self)
        self.new = self.modified = self.removed = self.unmodified = self.bundlediff = None

    def start(self, bundles=None,
              new=False, modified=True, removed=True, unmodified=False,
              dirs=None, bundlediff=False):
        if not bundles:
            bundles = []
        else:
            new = False
        if dirs == None:
            dirs = [self.wsPath]

        assert not (new and bundlediff), "mutually exclusive"

        self.new, self.modified, self.removed, self.unmodified = new, modified, removed, unmodified
        self.bundlediff = bundlediff

        self._validateDirs(dirs)

        root = os.path.join(self.wsMetaDir, "info")
        workspace_files = set()
        md5_files = [os.path.splitext(f)[0] for f in os.listdir(root) if f.endswith(".md5sum")]
        # If we are filtering with bundles check that all exist
        for bundle in bundles:
            if bundle not in md5_files:
                raise InvalidUserInput("Invalid bundle name '%s'" % bundle)
        if bundles:
            md5_files = bundles
        # Go through .md5sum files
        for md5name in md5_files:
            with open(os.path.join(root, md5name + ".md5sum")) as f:
                # Check all files named in md5sum file
                for line in f:
                    (pkg_md5, _, pkg_file) = line.partition("  ")
                    pkg_file = os.path.normcase(pkg_file)[:-1]
                    if new:
                        # Use md5 to save memory
                        workspace_files.add(md5(pkg_file).digest())
                    if modified or removed or unmodified:
                        skip = False
                        pkg_file = os.path.join(self.wsPath, pkg_file)
                        if pkg_file.startswith(tuple(dirs)):
                            try:
                                with open(pkg_file, "rb") as pf:
                                    if pkg_md5 != md5(pf.read()).hexdigest():
                                    # TODO: Could be optimized with checking mod time or
                                    # by creating new processes for calculating md5
                                    #if modified and os.path.getmtime(pkg_file) == 0:
                                        if modified:
                                            skip = self.output("M", md5name, pkg_file)
                                    elif unmodified:
                                        skip = self.output("U", md5name, pkg_file)
                            except IOError:
                            #except OSError, e:
                                if removed:
                                    skip = self.output("R", md5name, pkg_file)
                        if skip:
                            break
        self.output("END", None, None)
        if new:
            self._findNewFiles(dirs, workspace_files)

    def _validateDirs(self, dirs):
        for i, d in enumerate(dirs):
            dirs[i] = os.path.normcase(os.path.abspath(d))
            if not dirs[i].startswith(self.wsPath) or not os.path.isdir(dirs[i]):
                raise InvalidUserInput("Search path '%s' not found from under current workspace" % dirs[i])

        # Check that dirs do not include other dirs
        for i, d in enumerate(dirs):
            for d2 in dirs[i+1:]:
                if not utils.pathsUnique(d, d2):
                    raise InvalidUserInput("Search paths '%s' and '%s' are in conflict. Please remove either one." % (d, d2))

    def _findNewFiles(self, dirs, existing_files):
        for d in dirs:
            for root, dirs, files in os.walk(d):
                for name in files:
                    rel_path = os.path.normcase(os.path.join(root[len(self.wsPath):], name))
                    if rel_path[0] == os.sep:
                        rel_path = rel_path[1:]
                    rel_path = md5(rel_path).digest()
                    if rel_path not in existing_files:
                        self.output("N", None, os.path.normcase(os.path.join(root, name)))