#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Default processors
#

import os
import re
import xml.sax

from Blocks.Packaging.PackageWriter import Dependency
from Blocks.Packaging.Logging import Logging

class BuildDataDependencyProcessor(object):
    '''
    Call this class with a Component as argument to add dependencies from a
    BuildData object. Normally this is the same data that was used to make the
    component in the first place.

    Indirect dependencies are handled here, by applying rules to source package
    dependencies.
    '''

    def __init__(self, indirectDependencyRules):
        '''
        Reads dependency data from buildData and applies it to the component's
        member packages. BuildData is expected to yield file->file dependency
        data when the buildData.getDependencies() is called.

        If a configuration is specified, indirect dependency rules are applied
        to the component. The configuration can be passed as one of:
            - The path to the configuration file
            - A file-like object from which the configuration can be read
            - A List containing a previously parsed configuration

        @param indirectDependencyRules: Configuration
        @type indirectDependencyRules: None, String, file-like object or List
        '''
        self.globalFileMap = None
        if indirectDependencyRules:
            if isinstance(indirectDependencyRules, list):
                # previously parsed ruleset
                self.indirectDependencyRules = indirectDependencyRules
            else:
                # string containing path
                self.indirectDependencyRules = IndirectDependencyRules(indirectDependencyRules).getRules()
        else:
            self.indirectDependencyRules = None
        self.logger = Logging.getLogger("pfw.processors.builddata")

    def __call__(self, component, buildData, globalFileMap):
        self.globalFileMap = globalFileMap
        if component.getName() != buildData.componentName:
            raise ValueError("Bad dependency info for '%s'. BuildData is for another component '%s'." %
                (component.getName(), buildData.componentName))

        for dble in buildData.getDependencies():
            # dble is a file belonging to this component (BuildData.BdFile)
            # with dependencies on other files.
            ownerPackageName = self.globalFileMap.getPackageName(dble.getPath())
            if not ownerPackageName:
                self.logger.warning("Dependency data includes file '%s', not in filemap.", dble.getPath())
                continue
            ownerPackage = component.getPackage(ownerPackageName)
            if not ownerPackage:
                self.logger.warning("Dependency data includes file '%s', not in the component '%s'", dble.getPath(), component.getName())
                continue

            for dependency in dble.getOwnerDependencies():
                self.fileDep(ownerPackage, dependency)

            for dependency in dble.getSourceDependencies():
                if component.getSourcePackage():
                    self.fileDep(component.getSourcePackage(), dependency)
                else:
                    self.logger.warning("Cannot add dependency to %s in non-existent source package (%s)", dependency, component.getName())

            if self.indirectDependencyRules:
                for library in dble.getSourceDependencies():
                    library = os.path.basename(library)    # rules only use filename
                    for matchExp, pkgExp, fileExp in self.indirectDependencyRules:
                        if matchExp.match(library):
                            for exp, depType in fileExp:
                                if dble.variantPlatform:
                                    exp = exp.replace("$(VARIANTPLATFORM)", dble.getVariantPlatform())
                                if dble.variantType:
                                    exp = exp.replace("$(VARIANTTYPE)", dble.getVariantType())
                                self.fileDep(ownerPackage, exp, depType)
                            for exp, depType in pkgExp:
                                if dble.variantPlatform:
                                    exp = exp.replace("$(VARIANTPLATFORM)", dble.getVariantPlatform())
                                if dble.variantType:
                                    exp = exp.replace("$(VARIANTTYPE)", dble.getVariantType())
                                if ownerPackage.addDependency(Dependency(exp, type=depType)):
                                    self.logger.debug("Indirect dependency (file): %s now depends on %s",
                                        ownerPackage.getName(), Dependency(exp, type=depType))

    def fileDep(self, package, fileName, depType="strong"):
        depName = self.globalFileMap.getPackageName(fileName)
        if depName:
            if package.addDependency(Dependency(depName, type=depType)):
                self.logger.debug("%s now depends on %s", package.getName(), depName)
        else:
            self.logger.warning("Unable to find package owning dependency '%s'", fileName)

class IndirectDependencyRules(object):
    ''' Indirect dependency rules '''
    class ConfigReader(xml.sax.ContentHandler):
        '''
        Read the configuration and store the results in self.rules, which is a
        list of tuples: ("<match>", [<pkgrules>], [<filerules>]).

        Each <pkgrule> and <filerule> is composed of a ("strong|weak", "path")
        tuple. The path may contain variables, currently supported:
            - $(VARIANTPLATFORM)
            - $(VARIANTTYPE)
        '''
        def __init__(self):
            self.rules = []
            self.matchExp = ""
            self.packageExp = []
            self.fileExp = []
            self.cdata = []
            self.cdString = ""
            self.depType = None

        def characters(self, data):
            self.cdata.append(data)

        def startElement(self, tag, attributes):
            self.cdata = []
            if tag == "depends":
                self.depType = attributes.get("type", "weak").encode("ascii")

        def endElement(self, tag):
            self.cdString = "".join(self.cdata).strip()
            self.cdata = []
            if tag == "library":
                self.matchExp = re.compile(self.cdString)
            elif tag == "file":
                self.fileExp.append((self.cdString, self.depType))
            elif tag == "package":
                self.packageExp.append((self.cdString, self.depType))
            elif tag == "dep":
                self.rules.append((self.matchExp, self.packageExp, self.fileExp))
                self.matchExp = ""
                self.packageExp = []
                self.fileExp = []

    def __init__(self, configuration):
        '''

        @param configuration: Configuration for indirect dependencies
        @type configuration: Path to file or file-like object
        '''
        p = xml.sax.make_parser()
        i = IndirectDependencyRules.ConfigReader()
        p.setContentHandler(i)
        if hasattr(configuration, "read"):
            f = configuration
            f.seek(0)
        else:
            f = open(configuration)
        while True:
            data = f.read()
            if not data:
                break
            p.feed(data)
        p.close()
        self.rules = i.rules

    def getRules(self):
        return self.rules
