#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Package writer
#

'''
Writer classes create packages, others describe the interface for passing
package data to the writers.
@todo: zip and rpm writers are not thread safe.
'''

import os
import re
import copy
from math import ceil
import textwrap
import time
from zipfile import ZipFile, ZIP_DEFLATED
from StringIO import StringIO
import tarfile

try:
    from hashlib import md5
except ImportError:
    import md5

from Blocks.Packaging.Logging import Logging
from Blocks import arfile
from Blocks import debfile

class VersionInfo(object):
    '''
    Bestows ability to set and get version and build IDs when inherited.

    Used by PackageWriter when handling Packageables and Filesets.
    '''
    def __init__(self, version=None, buildId=None, revision=None):
        self.__identifier = Identifier(version, buildId, revision)

    def getIdentifier(self):
        '''
        @return: The identifier - version and build Id info
        @rtype: Identifier
        '''
        return self.__identifier

    def setIdentifier(self, id):
        '''
        @param id: A version string or Identifier instance
        @type id: String or Identifier
        '''
        if isinstance(id, str):
            self.__identifier = Identifier(id)
        elif isinstance(id, Identifier):
            self.__identifier = id
        else:
            raise TypeError, "Identifier must be a string or Identifier"

    def setVersion(self, version):
        '''
        Convenience method
        @param version: The version component of the identifier:
        @type version: String
        '''
        self.getIdentifier().setVersion(version)

    def getVersion(self):
        '''
        Convenience method
        @return: The version component of the identifier
        @rtype: String
        '''
        return self.getIdentifier().getVersion()

    def setBuild(self, id):
        '''
        Convenience method
        @param id: The build component of the identifier
        @type id: String
        '''
        self.getIdentifier().setBuild(id)

    def getBuild(self):
        '''
        Convenience method
        @return: The build component of the identifier
        @rtype: String
        '''
        return self.getIdentifier().getBuild()

    def setRevision(self, revision):
        '''
        Convenience method
        @param revision: The revision component of the identifier
        @type revision: String
        '''
        self.getIdentifier().setRevision(revision)

    def getRevision(self):
        '''
        Convenience method
        @return: The revision component of the identifier
        @rtype: String
        '''
        return self.getIdentifier().getRevision()

    def setRelease(self, value):
        '''
        Convenience method
        @param value: The release component of the identifier
        @type value: String
        '''
        self.getIdentifier().setRelease(value)

    def getRelease(self):
        '''
        Convenience method
        @return: The release component of the identifier
        @rtype: String
        '''
        return self.getIdentifier().getRelease()

class Packageable(VersionInfo):
    '''
    The top level class in the PackageWriter interface. A collection of
    FileSets.

    Packageable, FileSet, Dependency and VersionInfo define the interface for
    PackageWriter and implement some basic functionality.
    '''
    def __init__(self):
        VersionInfo.__init__(self, "1", "1", None)
        self.__packageAllFlag = True
        self.__sourcePackage = None

    def getPackages(self):
        '''
        @return: List(FileSet)
        '''
        raise NotImplementedError

    def getPackage(self):
        '''
        @return: FileSet
        '''
        raise NotImplementedError

    def getName(self):
        '''
        @return: the name
        @rtype: String
        '''
        raise NotImplementedError

    def getPackageAllFlag(self):
        '''
        @return: packageAll
        @rtype: Boolean
        '''
        return self.__packageAllFlag

    def setPackageAllFlag(self, value):
        '''
        @param value: Should all packages be packaged
        @type value: Boolean
        '''
        if type(value) != bool:
            raise TypeError, "packageAllFlag must be a Boolean"
        self.__packageAllFlag = value

    def getAttributes(self):
        '''
        Metadata excluding name and version. Can be empty.

        @return: Attributes
        @rtype: Dictionary(attribute)
        '''
        raise NotImplementedError

    def getAttribute(self, name):
        '''
        A convenience method to access attributes.
        @param name: attribute name
        '''
        if name in self.getAttributes():
            return self.getAttributes()[name]
        else:
            return None

    def setSourcePackage(self, package):
        if not isinstance(package, FileSet):
            raise TypeError("Source package must be a FileSet. Got '%s'" % type(package))
        self.__sourcePackage = package

    def getSourcePackage(self):
        return self.__sourcePackage

class FileSet(VersionInfo):
    '''
    An interface class for PackageWriter; stores filenames.

    See Package for the concrete implementation.
    '''
    def __init__(self):
        VersionInfo.__init__(self, "1", "1")
        self.__packageFlag = True
        self.__lastPackagedVersion = Identifier()
        self.__apiVersion = None

    def getFiles(self):
        '''
        @return: The
        @rtype: List(path)
        '''
        raise NotImplementedError

    def getName(self):
        '''
        @return: The name of the fileset
        '''
        raise NotImplementedError

    def getPackageFlag(self):
        '''
        @return: packageFlag - Should the package be written
        @rtype: Boolean
        '''
        return self.__packageFlag

    def setPackageFlag(self, value):
        '''
        @param value: Should the package be written
        '''
        if type(value) != bool:
            raise TypeError, "packageFlag must be a Boolean"
        self.__packageFlag = value

    def getDependencies(self):
        '''
        @return: Dependencies that describe what other filesets this depends on
        @rtype: List(Dependency)
        '''
        raise NotImplementedError

    def setLastPackagedVersion(self, last):
        '''
        Store the previously packaged version
        @param last: A FileSet having the last packaged version, or None to
            use the current package version, or an Identifier.
        @type last: FileSet or None or Identifier
        '''
        if isinstance(last, FileSet):
            if last is self:
                self.__lastPackagedVersion = copy.deepcopy(self.getIdentifier())
            else:
                self.__lastPackagedVersion = last.getLastPackagedVersion()
        elif last == None:
            self.__lastPackagedVersion = copy.deepcopy(self.getIdentifier())
        elif isinstance(last, Identifier):
            self.__lastPackagedVersion = last
        else:
            raise TypeError, "setLastPackagedVersion takes a FileSet, an Identifier or None as parameter"

    def getLastPackagedVersion(self):
        '''
        @return: version
        @rtype: String
        '''
        return self.__lastPackagedVersion

    def setApiVersion(self, version):
        '''
        Store the API version

        @param version: The version of the package API
        @type version: String
        '''
        if type(version) != str:
            raise TypeError, "API version must be a string"
        try:
            major, minor = version.split(".", 1)
            assert major.isdigit()
            assert minor.isdigit()
            assert (int(major) > 0)
            assert (int(minor) >= 0)
        except Exception:
            raise AttributeError("API version must be a dot-separated pair of integers")
        self.__apiVersion = version

    def getApiVersion(self):
        '''
        @return: API version
        @rtype: String or None
        '''
        return self.__apiVersion

    def getApiFingerprint(self):
        '''
        @return: API Fingerprint
        @rtype: String
        '''
        raise NotImplementedError

class Dependency(VersionInfo):
    ''' Describes the dependency relation of a package to another '''
    def __init__(self, package, type="strong", version=None, buildId=None, eval=None, revision=None):
        '''
        By default the dependency defines a strong dependency on a package. The
        version, build and evaluator have to be explicitly set.

        @param package: The name of the package required
        @type package: String
        @param type: Either 'strong' or 'weak'
        @type type: String
        @param version: The version of the required package.
        @type version: String
        @param buildId: The build ID of the required package
        @type buildId: String
        @param eval: Evaluator to define an exact version or a range of versions
        @type eval: String
        '''
        VersionInfo.__init__(self, version=version, buildId=buildId, revision=revision)
        self.package = package
        self.setEval(eval)
        self.setType(type)

    def __str__(self):
        if self.eval:
            e = self.eval + " " + str(self.getIdentifier())
        else:
            e = ""
        if self.type == "weak":
            return "weak(%s %s)" % (self.package, e)
        else:
            return "strong(%s %s)" % (self.package, e)

    def __eq__(self, other):
        '''
        @param other:
        @type other:
        '''
        if self.package == other.package and \
            self.eval == other.eval and \
            self.type == other.type and \
            self.getIdentifier() == other.getIdentifier():
            return True
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def rpmString(self):
        ''' Return RPM-friendly "Requires" string '''
        if self.eval:
            if self.eval == "==":
                e = " = " + str(self.getIdentifier())
            else:
                e = " %s %s" % (self.eval, self.getIdentifier().toString())
        else:
            e = ""
        if self.type == "weak":
            return "Requires(hint): %s%s" % (self.package, e)
        else:
            return "Requires: %s%s" % (self.package, e)

    def debString(self):
        ''' Return Deb-friendly string '''
        if self.eval:
            if self.eval == "==":
                e = "="
            elif self.eval == "<":
                e = "<<"
            elif self.eval == ">":
                e = ">>"
            else:
                e = self.eval
            return "%s (%s %s)" % (self.package, e, self.getIdentifier().toString())
        else:
            return self.package

    def setType(self, type):
        '''
        @param type: strong|weak
        '''
        if type in ("strong", "weak"):
            self.type = type
        elif type == "":
            self.type = "strong"
        else:
            raise ValueError("Not a valid dependency type: '%s'" % type)

    def setEval(self, eval):
        '''
        @param eval: "==" | "<=" | ">=" | "<" | ">" | None
        '''
        if eval in ("==", "<=", ">=", "<", ">", None):
            self.eval = eval
        else:
            raise ValueError("Not a valid dependency evaluator: '%s'" % eval)

    def getType(self):
        return self.type

    def getPackage(self):
        return self.package

    def getEval(self):
        return str(self.eval)

class Identifier(object):
    '''
    A container for version and build information, replacing a simple version
    string.

    The Identifier consists of three units: version, revision and build. Version
    and revision make up the externally assigned identifier. Normally these are
    concatenated to make the identifier string: VERSION-REVISION.

    The build ID identifies a package in the build history, and is thus internal
    to the packaging framework. It can, however feature in the identifier string
    in two ways:
        - the revision can contain "$b", which is replaced by the build ID
        - if setUseEpoch() is called it becomes the epoch, which overrides
          the version string in priority. In this case the identifier string
          becomes EPOCH:VERSION-REVISION.

    Epoch should only be used when the version is not guaranteed to increment in
    subsequent builds. Epoch can only have the value of the build ID.
    '''

    VERSION_RE = re.compile("^\w[\w\.~+-]*$")
    REVISION_RE = re.compile("^\w|\$b|\$c[\w\.~+]|\$b|\$c*$")

    def __init__(self, version=None, build=None, revision=None):
        '''
        @param version: Version
        @type version: String
        @param build: Build
        @type build: String
        '''
        self.__version = None
        self.__build = None
        self.__revision = None
        self.__releaseCounter = "1"
        self.__useEpoch = False
        self.setVersion(version)
        self.setBuild(build)
        self.setRevision(revision)

    def getVersion(self):
        '''
        @return: Version
        @rtype: String
        '''
        return self.__version

    def setVersion(self, version):
        '''
        Allowed values are alphanumeric characters and .~+-. The first character
        must be alphanumeric.

        @param version: Version
        @type version: None or String
        '''
        if version != None:
            if not isinstance(version, basestring):
                raise TypeError("Version must be a string or None; got '%s'" % type(version))
            if not Identifier.VERSION_RE.match(version):
                raise ValueError("Empty version or illegal characters in version string: %s" % version)
            if "\n" in version:
                raise ValueError("Cannot have linebreaks in version string")
        self.__version = version

    def getBuild(self):
        '''
        @return: Build ID
        @rtype: String
        '''
        return self.__buildId

    def setBuild(self, id):
        '''
        Build must be a string containing digits.

        @param id: Build ID
        @type id: String
        '''
        if id is not None:
            if not isinstance(id, basestring):
                raise TypeError("buildId must be a string or None")
            if not id.isdigit():
                raise ValueError("buildId must be a string containing digits")
            if not id:
                raise ValueError("buildId cannot be an empty string")
        self.__buildId = id

    def getRevision(self):
        '''
        @return: Revision
        @rtype: String
        '''
        if self.__revision:
            return self.__revision.replace("$b", self.getBuild() or "").replace("$c", str(self.__releaseCounter))
        else:
            return self.__releaseCounter

    def setRevision(self, revision):
        '''
        Allowed values are alphanumeric characters and .~+. The first character
        must be alphanumeric.

        The exception is "$b", which can occur anywhere and is replaced by the
        build ID.

        @param revision: Revision
        @type revision: String
        '''
        if revision != None:
            if not isinstance(revision, basestring):
                raise TypeError("Revision must be a string or None; got '%s'" % type(revision))
            if not Identifier.REVISION_RE.match(revision):
                raise ValueError("Empty version or illegal characters in revision string: %s" % revision)
            if "\n" in revision:
                raise ValueError("Cannot have linebreaks in revision string")
        self.__revision = revision

    def toString(self, format=None):
        ''' Return the identifier with optional format.

        If format is specified, it is returned with the following replacements::

            $b -> build
            $v -> version
            $r -> revision
            $c -> release counter

        If format is not specified, the default format is used:
            - normally I{version[-revision]}
            - If build is epoch: I{[epoch:]version[-revision]}

        "$b" is always replaced by the build in the revision string.
        "$c" is always replaced by the release counter in the revision string.

        @param format: Format string
        @type format: String
        '''
        if format:
            format = format.replace("$b", self.getBuild() or "")
            format = format.replace("$v", self.getVersion() or "")
            format = format.replace("$r", self.getRevision() or "")
            format = format.replace("$c", self.getRelease() or "")
            return format
        else:
            out = self.getVersion()
            if self.getUseEpoch() and self.getBuild() != "0":
                out = "%s:%s" % (self.getBuild(), out)
            if self.getRevision():
                out += "-%s" % self.getRevision()
            return out

    def setRelease(self, value):
        assert value.isdigit()
        self.__releaseCounter = value

    def getRelease(self):
        return self.__releaseCounter

    def setUseEpoch(self, value=None):
        if value != False:
            value = True
        self.__useEpoch = value

    def getUseEpoch(self):
        return self.__useEpoch

    def getEpoch(self):
        if self.getUseEpoch():
            return self.getBuild()
        else:
            return "0"

    def __eq__(self, other):
        return (isinstance(other, Identifier) and
            self.getBuild() == other.getBuild() and
            self.getVersion() == other.getVersion() and
            self.getRevision() == other.getRevision())

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self.toString(format="$b:$v:$r")

    @staticmethod
    def fromString(string):
        ''' Deserialize from str(Identifier) '''
        build, version, revision = string.split(":")
        return Identifier(version or None, build or None, revision or None)

class PackageWriter(object):
    '''
    Abstract class.

    A PackageWriter packages files from the filesystem into a set of one or more
    packages.

    The classes Packageable.FileSet and Packageable.Packageable define how
    data is passed to a PackageWriter. Files are listed in FileSets, which are
    are passed to a PackageWriter in a Packageable.
    '''
    def __init__(self, options=None):
        self.packageable = None
        self.options = options
        self.logger = Logging.getLogger("pfw.packagewriter")

    def write(self, packageable, epocRoot, destinationDir, sourceRoot):
        '''
        Writes package file(s) from a Packageable

        @param packageable: The object (usually Component) to write into bundles.
        @type packageable: Packageable
        @param epocRoot: The path to the directory containg epoc32.
        @type epocRoot: String
        @param destinationDir: The path where packages should be written.
        @type destinationDir: String
        @param sourceRoot: The path to the directory containing src.
        @type sourceRoot: String
        '''
        written = []
        self.setPackageable(packageable)
        destinationDir = os.path.normpath(destinationDir)
        if not os.path.isdir(destinationDir):
            raise ValueError("Not a directory: %s" % destinationDir)

        if sourceRoot and packageable.getSourcePackage():
            if self.packageable.getPackageAllFlag() or packageable.getSourcePackage().getPackageFlag():
                written.append(self.writeSinglePackage(packageable.getSourcePackage(), sourceRoot, destinationDir))
        else:
            self.logger.info("%s - Not writing source bundle: sourcelist or sourceroot missing"%packageable.getName())

        if bool(packageable.getPackages()) ^ bool(epocRoot):
            self.logger.info("%s - Not writing target bundles: packages or epocroot missing"%packageable.getName())
        else:
            for p in self.packageable.getPackages():
                if self.packageable.getPackageAllFlag() or p.getPackageFlag():
                    out = self.writeSinglePackage(p, epocRoot, destinationDir)
                    if type(out) == tuple:
                        written += out
                    else:
                        written.append(out)
                else:
                    self.logger.debug("%s - Not writing %s: unchanged." % (packageable.getName(), p.getName()))
        return written

    def setPackageable(self, packageable):
        if not isinstance(packageable, Packageable):
            raise TypeError("PackageWriter only handles Packageables")
        self.packageable = packageable

    def writeSinglePackage(self, package, sourceDir, destinationDir):
        raise NotImplementedError("Cannot use abstract PackageWriter")

class ZipWriter(PackageWriter):
    ''' Write packages as compressed Zip files. '''
    def __init__(self, options=None):
        PackageWriter.__init__(self, options)

    def writeSinglePackage(self, package, sourceDir, destinationDir):
        pkgfile = os.path.join(destinationDir, package.getName() + ".zip")
        out = ZipFile(pkgfile, "w", compression=ZIP_DEFLATED)
        os.chdir(sourceDir)
        for p in package.getFiles():
            out.write(p.path)
        out.close()
        return pkgfile

class RpmWriter(ZipWriter):
    '''
    Overview
    ========
        Writes the ingredients to create RPM packages from a Packageable.
            - Zip source files
            - Specfiles
            - A makefile for creating the actual RPMs.

    Attributes
    ==========
        If the metadata required by rpmbuild is not available from Packageable or
        Fileset attributes, a warning is logged and dummy values are used.

        Below are lists of attributes used. Unlisted attributes are ignored.

        Mandatory Packageable attributes
        --------------------------------
            - summary
            - license
            - vendor
            - group
            - description

        Optional Packageable attributes
        -------------------------------
            - url
    '''

    def __init__(self, options=None):
        ZipWriter.__init__(self, options)
        self.specFileName = None
        self.specFile = None
        self.logger = Logging.getLogger("pfw.packagewriter.rpm")
        self.mandatoryPackageableAttributes = [
            #PackageableAttributeName, SpecAttributeName, default
            ("summary", "Summary", "-"),
            ("license", "License", "-"),
            ("vendor", "Vendor", "-"),
            ("group", "Group", "devel")]
        self.mandatorySectionNames = [("description", "description")]
        self.optionalPackageableAttributes = [("url", "URL")]

    def writeSinglePackage(self, package, sourceDir, destinationDir):
        relocatePrefix = "/usr/local/src"
        rpmSpecDir = os.path.join(destinationDir, "SPECS/")
        rpmSrcDir = os.path.join(destinationDir, "SOURCES/")
        if not os.path.isdir(rpmSpecDir):
            os.makedirs(rpmSpecDir)
        if not os.path.isdir(rpmSrcDir):
            os.makedirs(rpmSrcDir)

        zip = ZipWriter.writeSinglePackage(self, package, sourceDir, rpmSrcDir)

        versionCharacters = re.compile(r"[^\.a-zA-Z0-9_]")
        if versionCharacters.search(package.getVersion()):
            raise ValueError("Unsuitable characters in package version %s" % package.getVersion())
        self.specFileName = rpmSpecDir + package.getName() + ".spec"
        self.specFile = open(self.specFileName, "wb") #we want unix linebreaks
        self.specFile.write("%define __os_install_post %{nil}\n")
        self.specFile.write("%define __find_requires   /bin/true\n")
        self.specFile.write("Name: " + package.getName() + "\n")
        self.specFile.write("Source: " + rpmSpecDir + "%{name}.zip\n")
        self.specFile.write("Buildroot: %{_tmppath}/%{name}-buildroot\n")
        self.specFile.write("BuildArch: noarch\n")
        self.specFile.write("Version: %s\n" % package.getVersion())
        self.specFile.write("Release: %s\n" % package.getRevision())
        #self.specFile.write("Epoch: %s\n"%package.getBuild())
        self.specFile.write("Prefix: %s\n" % relocatePrefix)

        for n, s, d in self.mandatoryPackageableAttributes:
            if n in self.packageable.getAttributes():
                self.specFile.write(s + ": " + str(self.packageable.getAttribute(n)) + "\n")
            else:
                self.logger.info("Component attribute '%s' is not defined, using '%s'", n, d)
                self.specFile.write("%s: %s\n" % (s, d))

        for k, v in self.optionalPackageableAttributes:
            if k in self.packageable.getAttributes():
                self.specFile.write(v + ": " + str(self.packageable.getAttribute(k)) + "\n")

        for d in package.getDependencies():
            self.specFile.write(d.rpmString()+"\n")

        if self.packageable.getAttribute("description"):
            description = self.packageable.getAttribute("description")
        else:
            description = "NO DESCRIPTION"
        self.specFile.write("\n%description\n")
        self.specFile.write("%s\n\n"%description)

        self.specFile.write("%prep\n")
        self.specFile.write("%setup -c -n %{name}\n")

        self.specFile.write("\n%install\n")
        self.specFile.write("rm -rf $RPM_BUILD_ROOT%s\n"%relocatePrefix)
        self.specFile.write("mkdir -p $RPM_BUILD_ROOT%s\n"%relocatePrefix)
        self.specFile.write("cp -r * $RPM_BUILD_ROOT%s/\n"%relocatePrefix)

        self.specFile.write("\n%files\n%defattr(644,root,root)\n")
        for f in package.getFiles():
            the_path = relocatePrefix + "/" + f.getPath()
            self.specFile.write(the_path + "\n")

        self.specFile.close()
        return (zip, self.specFileName)

class DebWriter(PackageWriter):
    '''
    Overview
    ========
       Writes deb archives from Packageables.

    Attributes
    ==========
        Each Deliverable belonging to the Packageable must have a size.

        Recommended Packageable attributes that should be set are:
            - group
            - summary
            - description

        The summary and description are used to create the deb Description, and
        as such the summary should be brief - it is truncated to 60 characters.
        The description is meant to be a longer text.

        Unlisted attributes are also included. The attribute name is
        capitalized and a prefix "X-" is added if necessary.
    '''

    DEP_MAPPINGS = ("Depends", "Enhances", "Recommends", "Suggests", None)

    def __init__(self, options=None):
        PackageWriter.__init__(self, options)
        self.logger = Logging.getLogger("pfw.packagewriter.deb")

    @staticmethod
    def getFileTarInfo(name, file):
        '''
        Create TarInfo for file-like object

        @param name: Name for tar member file
        @param file: file-like object
        '''
        ti = tarfile.TarInfo()
        ti.name = name
        if hasattr(file, "len"):
            try:
                int(file.len)
                ti.size = file.len
            except ValueError:
                ti.size = 0
        else:
            try:
                ti.size = len(file)
            except TypeError:
                pass
        ti.mtime = long(time.time())
        ti.mode = int(0644)
        ti.uname = "user"
        ti.gname = "group"
        ti.uid = 0
        ti.gid = 0
        return ti

    @staticmethod
    def getDirTarInfo(path):
        '''
        For adding arbitrary directory paths to a tarfile
        @param path:
        @type path:
        '''
        ti = tarfile.TarInfo()
        ti.name = path
        ti.type = tarfile.DIRTYPE
        ti.mtime = long(time.time())
        ti.mode = int(0755)
        ti.uname = "user"
        ti.gname = "group"
        ti.uid = 0
        ti.gid = 0
        return ti

    def writeSinglePackage(self, package, sourceDir, destinationDir):
        revision = ""
        if package.getRevision():
            revision = "-%s" % package.getRevision()
        debFileName = "%s_%s%s.deb" % (package.getName(), package.getVersion(), revision)

        # copyright
        copyright_file = StringIO()
        if self.packageable.getAttribute("vendor"):
            v = "Vendor: " + self.packageable.getAttribute("vendor").strip()
            copyright_file.write(v[:80] + "\n")
        if "license" in self.packageable.getAttributes():
            l = "License: " + self.packageable.getAttribute("license")
            wrap = textwrap.TextWrapper(subsequent_indent=" ", width=80)
            copyright_file.write("\n .\n".join([wrap.fill(line) for line in l.split("\n")]))
            copyright_file.write("\n")
        copyright_file.write("\n")
        copyright_file.seek(0)

        # md5sum
        md5sum_file = StringIO()
        for p in package.getFiles():
            if hasattr(md5, "new"):
                m = md5.new()
            else:
                m = md5()
            f = open(os.path.join(sourceDir, p.path), "rb")
            m.update(f.read())
            f.close()
            md5sum_file.write(m.hexdigest() + "  " + p.path + "\n")
        md5sum_file.seek(0)

        # control
        control_file = StringIO(self.__getControl(package))
        deb = debfile.DebFile(os.path.join(destinationDir, debFileName), mode="w")
        deb.addfile(arfile.ArInfo("debian-binary"))
        deb.write("2.0\n")
        deb.addfile(arfile.ArInfo("control.tar.gz"))
        f = tarfile.TarFile.open(mode="w:gz", fileobj=deb)
        f.addfile(self.getFileTarInfo("control", control_file), control_file)
        f.addfile(self.getFileTarInfo("md5sum", md5sum_file), md5sum_file)
        f.addfile(self.getFileTarInfo("copyright", copyright_file), copyright_file)
        f.close()

        # data
        deb.addfile(arfile.ArInfo("data.tar.gz"))
        d = tarfile.TarFile.open(mode="w:gz", fileobj=deb)
        seenDirs = []
        for p in package.getFiles():
            absPath = os.path.join(sourceDir, p.path)
            parentDir = None
            nextDir = os.path.dirname(p.path)
            # dpkg appears to require that each directory is listed in the tarball
            while os.path.dirname(p.path) not in seenDirs:
                parentDir = os.path.split(nextDir)[0]
                if parentDir in seenDirs or parentDir == "":
                    seenDirs.append(nextDir)
                    if nextDir != "":
                        d.addfile(self.getDirTarInfo(nextDir))
                    nextDir = os.path.dirname(p.path)
                else:
                    nextDir = parentDir
            # add file to tarfile without sourceDir
            tInfo = d.gettarinfo(absPath, p.path)
            tInfo.uname = "user"
            tInfo.gname = "group"
            tInfo.uid = 0
            tInfo.gid = 0
            if tInfo.mode & 0100:
                tInfo.mode = int(0755)
            else:
                tInfo.mode = int(0644)
            thefile = open(absPath, "rb")
            d.addfile(tInfo, thefile)
            thefile.close()
        d.close()
        if self.options:
            if "SIGN_KEYRING" in self.options and os.path.isdir(self.options.get("SIGN_KEYRING")):
                signType = self.options.get("SIGN_TYPE", "origin")
                try:
                    deb.addSignature(signType, self.options.get("SIGN_KEYRING"), True)
                except debfile.DebError, e:
                    self.logger.warning(str(e))
        deb.close()
        return debFileName

    def __getControl(self, fileset):
        '''Create a control file from Packageable and FileSet information

        @param fileset: The FileSet we are creating a Deb from
        @type fileset: Fileset
        @return: Text for control file
        @rtype: String
        '''
        installedSize = 0
        for p in fileset.getFiles():
            installedSize += int(p.size)

        # size is kbytes, let's round up to the nearest kB
        if installedSize > 0:
            installedSize = int(ceil(installedSize / 1024.0))

        control = ("Package: %s\n"
                   "Version: %s\n"
                   "Priority: optional\n"
                   "Installed-Size: %s\n"
                   "Architecture: all\n") % (fileset.getName(), fileset.getIdentifier().toString(), installedSize)

        section = fileset.group or self.packageable.getAttribute("group") or None
        if section:
            control += "Section: %s\n" % section

        strongDep = "Depends"
        weakDep = "Recommends"
        if self.options:
            # None is a valid value, so the default has to be something else
            option = self.options.get("STRONG_DEP_MAPPING", "illegal")
            if option in self.DEP_MAPPINGS:
                strongDep = option
            option = self.options.get("WEAK_DEP_MAPPING", "illegal")
            if option in self.DEP_MAPPINGS:
                weakDep = option

        # If it's None we don't want any deps listed
        if strongDep:
            strongDepList = ", ".join([d.debString().strip() for d in fileset.getDependencies() if d.getType() == "strong"])
            if strongDepList:
                control += "%s: %s\n" % (strongDep, strongDepList)

        if weakDep:
            weakDepList = ", ".join([d.debString().strip() for d in fileset.getDependencies() if d.getType() == "weak"])
            if weakDepList:
                control += "%s: %s\n" % (weakDep, weakDepList)

        # TODO: This is actually package. Change name
        if fileset.replaces:
            packages = ", ".join(fileset.replaces)
            for field in ("Conflicts", "Replaces", "Provides"):
                control += "%s: %s\n" % (field, packages)

        for k, v in self.packageable.getAttributes().items():
            if k not in ("summary", "description", "group", "vendor", "license"):
                if k.startswith("X-"):
                    control += "%s: %s\n" % (k.capitalize(), v)
                else:
                    control += "X-%s: %s\n" % (k.capitalize(), v)

        description = ""
        if self.packageable.getAttribute("summary"):
            description += self.packageable.getAttribute("summary")[:60] + "\n"
        if self.packageable.getAttribute("description"):
            if not description:
                description = "\n"
            wrap = textwrap.TextWrapper(initial_indent=" ", subsequent_indent=" ", width=80)
            description += "\n .\n".join([wrap.fill(line) for line in self.packageable.getAttribute("description").split("\n")])
        if not description:
            description = "n/a"
        control += "Description: " + description.rstrip()
        control += "\n\n"
        return control
