#
# Copyright (c) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
# This component and the accompanying materials are made available
# under the terms of "Eclipse Public License v1.0"
# which accompanies this distribution, and is available
# at the URL "http://www.eclipse.org/legal/epl-v10.html".
#
# Initial Contributors:
# Nokia Corporation - initial contribution.
#
# Contributors:
#
# Description:
# Blocks plugin for raptor
#

""" Extract linker output to link-<time>.log """

import os
import sys
import time
import re
import xml.sax
import xml.sax.saxutils
from getopt import getopt, GetoptError

import generic_path
import filter_interface
from raptor_meta import BldInfFile

class LinkInfoFilter(filter_interface.Filter):
    ''' Use of LinkInfoReader to collect output. '''

    def __init__(self):
        self.epocroot = str(generic_path.Path(os.environ.get("EPOCROOT", "")))
        assert self.epocroot, "Cannot proceed without EPOCROOT"
        assert os.path.isdir(self.epocroot), "EPOCROOT %s is not a directory" % self.epocroot

    def open(self, raptor_instance):
        """Open a log file for the various I/O methods to write to."""
        self.logDir = str(generic_path.Path(os.environ.get("FILTER_OUT_DIR", "")))
        if not (self.logDir and os.path.isdir(self.logDir)):
            self.logDir = str(generic_path.Path(os.environ.get("SBS_BUILD_DIR", "")))
        if not (self.logDir and os.path.isdir(self.logDir)):
            if os.path.isdir(os.path.join(self.epocroot, "epoc32", "build")):
                self.logDir = str(generic_path.Path(os.path.join(self.epocroot, "epoc32", "build")))
            else:
                self.logDir = str(generic_path.Path(self.epocroot))
        self._linkTimeStamp = time.strftime("%Y-%m-%d-%H-%M-%S")
        self._linkFile = os.path.join(self.logDir, "link." + self._linkTimeStamp + ".log")
        self._linkOut = open(self._linkFile, "wb")
        self._linkParser = xml.sax.make_parser()
        self._linkInfoReader = LinkInfoReader(self.epocroot)
        self._linkParser.setContentHandler(self._linkInfoReader)
        return True

    def write(self, text):
        """Write text into the log file"""
        try:
            self._linkParser.feed(text)
        except Exception, e:
            sys.stderr.write("LinkInfoReader failed: %s. Last: %s" % (e, text))
            return False
        return True

    def summary(self):
        """Write Summary"""
        sys.stdout.write("Writing linker data to %s\n"%self._linkFile)
        return True

    def close(self):
        """Close the log file"""
        try:
            self._linkParser.close()
            self._linkInfoReader.writeXml(out=self._linkOut)
            self._linkOut.close()
        except Exception:
            return False
        return True

class ArmlinkCmd(object):
    """
    Read SBS output log to extract link info

    Example::
        /opt/rvct-2.2.616/bin/armlink
            --diag_suppress 6331
            --bpabi
            --reloc
            --no_scanlib
            --datacompressor=off
            --debug
            --dll
            --split
            --rw-base 0x400000
            --symver_soname
            --soname 'balclient{000a0000}[101fb5dd].dll'
            --entry _E32Dll '/data/build/sos94_09wk04/epoc32/release/armv5/udeb/edll.lib(uc_dll_.o)'
            -o /data/build/sos94_09wk04/epoc32/release/armv5/udeb/balclient.dll.sym
            --symbols
            --list /data/build/sos94_09wk04/epoc32/release/armv5/udeb/balclient.dll.map
            --via /data/build/sos94_09wk04/epoc32/build/framework/c_1b392bd557300014/balclient_dll/armv5/udeb/balclient_udeb_objects.via
            /data/build/sos94_09wk04/epoc32/release/armv5/udeb/edllstub.lib
            /data/build/sos94_09wk04/epoc32/release/armv5/udeb/usrt2_2.lib
            /data/build/sos94_09wk04/epoc32/release/armv5/lib/euser.dso
            /data/build/sos94_09wk04/epoc32/release/armv5/lib/drtaeabi.dso
            /data/build/sos94_09wk04/epoc32/release/armv5/lib/dfpaeabi.dso
            /data/build/sos94_09wk04/epoc32/release/armv5/lib/dfprvct2_2.dso
            /data/build/sos94_09wk04/epoc32/release/armv5/lib/drtrvct2_2.dso
            '/opt/rvct-2.2.616/lib/armlib/h_t__uf.l(switch8.o)'
    """

    @staticmethod
    def getOpts(command):
        if not command:
            raise ValueError("Not a command: '%s'" % command)
        if not os.path.basename(command.split()[0]) in ("armlink", "armlink.exe"):
            raise ValueError("Not an armlink command '%s'" % command)
        try:
            # annoying space-separated value for --entry
            parts = command.strip().split()[1:]
            for i in range(len(parts)):
                if parts[i] == "--entry":
                    parts[i+1] = "%s %s" % (parts[i+1], parts[i+2])
                    del parts[i+2]
                    break
            opts, leftovers = getopt(parts,
                                "ho:",
                                ["help",
                                 "output=",
                                 "via=",
                                 "partial",
                                 "scatter=",
                                 "ro-base=",
                                 "rw-base=",
                                 "bestdebug",
                                 "datacompressor=",
                                 "nodebug",
                                 "debug",
                                 "entry=",
                                 "libpath",
                                 "userlibpath",
                                 "nolocals",
                                 "noremove",
                                 "callgraph",
                                 "feedback=",
                                 "info=",
                                 "map",
                                 "symbols",
                                 "xref",
                                 # these are not listed in -h but found in commands
                                 "diag_suppress=",
                                 "bpabi",
                                 "reloc",
                                 "no_scanlib",
                                 "split",
                                 "dll",
                                 "symver_soname",
                                 "soname=",
                                 "list="
                                 ])
            return (opts, leftovers)
        except GetoptError:
            return None

    @staticmethod
    def getLibraries(command):
        '''
        Yield libraries from armlink command line string.
        @param command: Armlink command
        @type command: String
        '''
        candidates = []
        # get rid of obvious shit
        parts = ArmlinkCmd.getOpts(command)
        if parts:
            candidates = parts[1]
        else:
            # fall back
            candidates = [part for part in command.split()[1:] if not part.startswith("-")]
        wanted = [re.compile(r"epoc32/release/armv5/lib/.+.dso$", re.IGNORECASE)]
        for file in candidates:
            for regex in wanted:
                match = regex.search(file)
                if match:
                    yield match.group()

    @staticmethod
    def getTarget(command):
        '''
        Get target (output) file from command line string.
        @param command:
        @type command:
        '''
        parts = ArmlinkCmd.getOpts(command)
        if parts:
            opts = parts[0]
            for k, v in opts:
                if k in ("-o", "--output"):
                    regex = re.compile(r"epoc32/release/armv5/[^/]+/.+.sym$", re.IGNORECASE)
                    match = regex.search(v)
                    if match:
                        return match.group()



class LinkInfoReader(xml.sax.ContentHandler):
    '''
    Read Raptor output and collect infs->targets->library dependencies and other
    associated bits.

    Link info can be spewed as XML using writeXml().

    Stores link info in a dictionary::
        self.infTargets = {"path/to.inf": [
                {'component': u'foo',
                 'platform': u'armv5',
                 'path': u'epoc32/foo/foo.dll',
                 'libraries': [u'epoc32/release/armv5/lib/foo.dso', u'epoc32/release/armv5/lib/bar.dso']
                }
            ]
        }
    '''

    def __init__(self, epocroot=None):
        self.epocroot = epocroot
        self.cdata = []
        self.infTargets = {}    # where the data ends up.
                                # keys inf paths, values target lists
        self.isLinkRecipe = False
        self.currentInfPath = ""
        self.currentTarget = ""
        self.currentImportLib = ""

    def characters(self, data):
        self.cdata.append(data)

    def startElement(self, tag, attributes):
        # The last tag was a link recipe, now expect cdata to contain the link
        # command string.
        if self.isLinkRecipe:
            cdString = "".join(self.cdata).strip()
            for line in [ line for line in cdString.split("\n") if line.startswith("+") ]:
                line = line.lstrip(" +")
                # Lots of junk could be included, for our purposes it can be ignored
                try:
                    for dso in ArmlinkCmd.getLibraries(line):
                        self.currentTarget["libraries"].append(dso)
                except ValueError:
                    pass
#                    self.logger.debug("Line in link recipe could not be parsed by ArmlinkCmd: %s" % line)
            self.infTargets[self.currentInfPath].append(self.currentTarget)
            self.isLinkRecipe = False
            self.currentInfPath = ""
            self.currentTarget = None
        self.cdata = []
        if tag == "recipe":
            recipeName = attributes.get("name")
            if recipeName == "linkandpostlink":
                self.isLinkRecipe = True
                self.currentInfPath = attributes.get("bldinf")
                target = attributes.get("target")
                # INF path is absolute, all others are relative to epocroot
                if self.currentInfPath not in self.infTargets:
                    self.infTargets[self.currentInfPath] = []
                if self.epocroot:
                    rootIndex = len(self.epocroot)
                else:
                    rootIndex = target.find("epoc32")
                    if rootIndex == -1:
                        raise ValueError, "Cannot guess epocroot from a path without 'epoc32': '%s'" % attributes.get("target")
                    self.epocroot = target[:rootIndex]
                # lstrip required - self.epocroot from init arg may not have trailing separator
                # Raptor always outputs Unix style paths -> hard coded separator in lstrip.
                path = target[rootIndex:].lstrip("/")
                if path.endswith(".sym"):
                    path = path[:-4]

                targetFileRoot = os.path.splitext(os.path.basename(target))[0]
                libFileRoot = os.path.splitext(os.path.basename(self.currentImportLib))[0]
                if targetFileRoot == libFileRoot:
                    self.currentImportLib = ""

                self.currentTarget = {
                                      "path": path,
                                      "component": attributes.get("component"),
                                      "platform": attributes.get("platform"),
                                      "libraries": [],
                                      "dso": self.currentImportLib[rootIndex:].lstrip("/")
                                      }
            elif recipeName == "importlibtarget":
                self.currentImportLib = attributes.get("target")

    def writeXml(self, out=sys.stdout, encoding="iso-8859-1"):
        generator = xml.sax.saxutils.XMLGenerator

        class LinkWriter(generator):
            def writeLinkData(self, infsAndTargets, epocroot):
                generator.startElement(self, u"linkdata", {"epocroot": epocroot})
                for inf, targets in infsAndTargets.items():
                    generator.startElement(self, u"inf", {"path": inf, "ddir": BldInfFile.outputPathFragment(inf)})
                    for target in targets:
                        generator.startElement(self, u"target", {"path": target["path"],
                                                                 "platform": target["platform"],
                                                                 "component": target["component"],
                                                                 "dso": target.get("dso", "")})
                        generator.startElement(self, u"libraries", {})
                        for lib in target["libraries"]:
                            generator.startElement(self, u"lib", {})
                            generator.characters(self, lib)
                            generator.endElement(self, u"lib")
                        generator.endElement(self, u"libraries")
                        generator.endElement(self, u"target")
                    generator.endElement(self, u"inf")
                generator.endElement(self, u"linkdata")
        l = LinkWriter(out, encoding)
        l.writeLinkData(self.infTargets, self.epocroot)
